"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const R = require("ramda");
const const_utils_1 = require("./const-utils");
var Expression;
(function (Expression) {
    function createColumnRef(column) {
        return { kind: 'ColumnRef', column };
    }
    Expression.createColumnRef = createColumnRef;
    function createTableColumnRef(table, column) {
        return { kind: 'TableColumnRef', table, column };
    }
    Expression.createTableColumnRef = createTableColumnRef;
    function createConstant(valueText) {
        return { kind: 'Constant', valueText };
    }
    Expression.createConstant = createConstant;
    function createParameter(index) {
        return { kind: 'Parameter', index };
    }
    Expression.createParameter = createParameter;
    function createUnaryOp(op, operand) {
        return { kind: 'UnaryOp', op, operand };
    }
    Expression.createUnaryOp = createUnaryOp;
    function createBinaryOp(lhs, op, rhs) {
        return { kind: 'BinaryOp', lhs, op, rhs };
    }
    Expression.createBinaryOp = createBinaryOp;
    function createTernaryOp(lhs, op, rhs1, rhs2) {
        return { kind: 'TernaryOp', lhs, op, rhs1, rhs2 };
    }
    Expression.createTernaryOp = createTernaryOp;
    function createExistsOp(subquery) {
        return { kind: 'ExistsOp', subquery };
    }
    Expression.createExistsOp = createExistsOp;
    function createInOp(lhs, op, subquery) {
        return { kind: 'InOp', lhs, op, subquery };
    }
    Expression.createInOp = createInOp;
    function createFunctionCall(funcName, argList) {
        return { kind: 'FunctionCall', funcName, argList };
    }
    Expression.createFunctionCall = createFunctionCall;
    function createArraySubQuery(subquery) {
        return { kind: 'ArraySubQuery', subquery };
    }
    Expression.createArraySubQuery = createArraySubQuery;
    function createCase(branches, else_) {
        return { kind: 'Case', branches, else_ };
    }
    Expression.createCase = createCase;
    function createTypeCast(lhs, targetType) {
        return { kind: 'TypeCast', lhs, targetType };
    }
    Expression.createTypeCast = createTypeCast;
    function walkSome(expr, elseVal, handlers) {
        switch (expr.kind) {
            case 'ColumnRef':
                return handlers.columnRef == null ? elseVal : handlers.columnRef(expr);
            case 'TableColumnRef':
                return handlers.tableColumnRef == null
                    ? elseVal
                    : handlers.tableColumnRef(expr);
            case 'Constant':
                return handlers.constant == null ? elseVal : handlers.constant(expr);
            case 'Parameter':
                return handlers.parameter == null ? elseVal : handlers.parameter(expr);
            case 'UnaryOp':
                return handlers.unaryOp == null ? elseVal : handlers.unaryOp(expr);
            case 'BinaryOp':
                return handlers.binaryOp == null ? elseVal : handlers.binaryOp(expr);
            case 'TernaryOp':
                return handlers.ternaryOp == null ? elseVal : handlers.ternaryOp(expr);
            case 'ExistsOp':
                return handlers.existsOp == null ? elseVal : handlers.existsOp(expr);
            case 'InOp':
                return handlers.inOp == null ? elseVal : handlers.inOp(expr);
            case 'FunctionCall':
                return handlers.functionCall == null
                    ? elseVal
                    : handlers.functionCall(expr);
            case 'ArraySubQuery':
                return handlers.arraySubQuery == null
                    ? elseVal
                    : handlers.arraySubQuery(expr);
            case 'Case':
                return handlers.case == null ? elseVal : handlers.case(expr);
            case 'TypeCast':
                return handlers.typeCast == null ? elseVal : handlers.typeCast(expr);
        }
    }
    Expression.walkSome = walkSome;
    function walk(expr, handlers) {
        switch (expr.kind) {
            case 'ColumnRef':
                return handlers.columnRef(expr);
            case 'TableColumnRef':
                return handlers.tableColumnRef(expr);
            case 'Constant':
                return handlers.constant(expr);
            case 'Parameter':
                return handlers.parameter(expr);
            case 'UnaryOp':
                return handlers.unaryOp(expr);
            case 'BinaryOp':
                return handlers.binaryOp(expr);
            case 'TernaryOp':
                return handlers.ternaryOp(expr);
            case 'ExistsOp':
                return handlers.existsOp(expr);
            case 'InOp':
                return handlers.inOp(expr);
            case 'FunctionCall':
                return handlers.functionCall(expr);
            case 'ArraySubQuery':
                return handlers.arraySubQuery(expr);
            case 'Case':
                return handlers.case(expr);
            case 'TypeCast':
                return handlers.typeCast(expr);
        }
    }
    Expression.walk = walk;
    function equals(a, b) {
        switch (a.kind) {
            case 'ColumnRef':
                // `tbl.col` and `col` in an expression context must point to
                // the same column. Otherwise the expression would be invalid
                // because of an unambiguous column reference.
                if (b.kind === 'TableColumnRef')
                    return a.column === b.column;
                if (a.kind !== b.kind)
                    return false;
                return a.column === b.column;
            case 'TableColumnRef':
                // `tbl.col` and `col` in an expression context must point to
                // the same column. Otherwise the expression would be invalid
                // because of an unambiguous column reference.
                if (b.kind === 'ColumnRef')
                    return a.column === b.column;
                if (a.kind !== b.kind)
                    return false;
                return a.table === b.table && a.column === b.column;
            case 'Constant':
                if (a.kind !== b.kind)
                    return false;
                return a.valueText === b.valueText; // TODO: This doesn't work in some cases, e.g. with '1.0' and '1.00'
            case 'Parameter':
                if (a.kind !== b.kind)
                    return false;
                return a.index === b.index;
            case 'UnaryOp':
                if (a.kind !== b.kind || a.op !== b.op)
                    return false;
                return equals(a.operand, b.operand);
            case 'BinaryOp':
                if (a.kind !== b.kind || a.op != b.op)
                    return false;
                return ((equals(a.lhs, b.lhs) && equals(a.rhs, b.rhs)) ||
                    (const_utils_1.isOperatorCommutative(a.op) &&
                        equals(a.lhs, b.rhs) &&
                        equals(a.rhs, b.lhs)));
            case 'TernaryOp':
                if (a.kind !== b.kind || a.op != b.op)
                    return false;
                return (equals(a.lhs, b.lhs) &&
                    equals(a.rhs1, b.rhs1) &&
                    equals(a.rhs2, b.rhs2));
            case 'ExistsOp':
                if (a.kind !== b.kind)
                    return false;
                return false; // TODO
            case 'InOp':
                if (a.kind !== b.kind)
                    return false;
                return false; // TODO
            case 'FunctionCall':
                if (a.kind !== b.kind)
                    return false;
                return (a.funcName === b.funcName &&
                    R.zip(a.argList, b.argList).every(([ap, bp]) => equals(ap, bp)));
            case 'ArraySubQuery':
                if (a.kind !== b.kind)
                    return false;
                return false; // TODO
            case 'Case':
                if (a.kind !== b.kind)
                    return false;
                return (R.zip(a.branches, b.branches).every(([ab, bb]) => equals(ab.condition, bb.condition) && equals(ab.result, bb.result)) &&
                    ((a.else_ !== null && b.else_ !== null && equals(a.else_, b.else_)) ||
                        (a.else_ === null && b.else_ === null)));
            case 'TypeCast':
                if (a.kind !== b.kind)
                    return false;
                return equals(a.lhs, b.lhs) && a.targetType == b.targetType;
        }
    }
    Expression.equals = equals;
})(Expression = exports.Expression || (exports.Expression = {}));
var SelectListItem;
(function (SelectListItem) {
    function createSelectListExpression(expression, as) {
        return { kind: 'SelectListExpression', expression, as };
    }
    SelectListItem.createSelectListExpression = createSelectListExpression;
    function createAllTableFields(tableName) {
        return { kind: 'AllTableFields', tableName };
    }
    SelectListItem.createAllTableFields = createAllTableFields;
    function createAllFields() {
        return { kind: 'AllFields' };
    }
    SelectListItem.createAllFields = createAllFields;
    function walk(item, handlers) {
        switch (item.kind) {
            case 'SelectListExpression':
                return handlers.selectListExpression(item);
            case 'AllTableFields':
                return handlers.allTableFields(item);
            case 'AllFields':
                return handlers.allFields(item);
        }
    }
    SelectListItem.walk = walk;
})(SelectListItem = exports.SelectListItem || (exports.SelectListItem = {}));
var TableRef;
(function (TableRef) {
    function create(schema, table) {
        return { kind: 'TableRef', schema, table };
    }
    TableRef.create = create;
})(TableRef = exports.TableRef || (exports.TableRef = {}));
var TableExpression;
(function (TableExpression) {
    function createTable(table, as) {
        return { kind: 'Table', table, as };
    }
    TableExpression.createTable = createTable;
    function createSubQuery(query, as) {
        return { kind: 'SubQuery', query, as };
    }
    TableExpression.createSubQuery = createSubQuery;
    function createCrossJoin(left, right) {
        return { kind: 'CrossJoin', left, right };
    }
    TableExpression.createCrossJoin = createCrossJoin;
    function createQualifiedJoin(left, joinType, right, condition) {
        return { kind: 'QualifiedJoin', left, joinType, right, condition };
    }
    TableExpression.createQualifiedJoin = createQualifiedJoin;
    function walk(tableExpr, handlers) {
        switch (tableExpr.kind) {
            case 'Table':
                return handlers.table(tableExpr);
            case 'SubQuery':
                return handlers.subQuery(tableExpr);
            case 'CrossJoin':
                return handlers.crossJoin(tableExpr);
            case 'QualifiedJoin':
                return handlers.qualifiedJoin(tableExpr);
        }
    }
    TableExpression.walk = walk;
})(TableExpression = exports.TableExpression || (exports.TableExpression = {}));
var OrderBy;
(function (OrderBy) {
    function create(expression, order, nulls) {
        return { expression, order, nulls };
    }
    OrderBy.create = create;
})(OrderBy = exports.OrderBy || (exports.OrderBy = {}));
var Limit;
(function (Limit) {
    function create(count, offset) {
        return { count, offset };
    }
    Limit.create = create;
})(Limit = exports.Limit || (exports.Limit = {}));
var SelectBody;
(function (SelectBody) {
    function create(selectList, from, where, groupBy, having) {
        return { selectList, from, where, groupBy, having };
    }
    SelectBody.create = create;
})(SelectBody = exports.SelectBody || (exports.SelectBody = {}));
var SelectOp;
(function (SelectOp) {
    function create(op, duplicates, select) {
        return { op, duplicates, select };
    }
    SelectOp.create = create;
})(SelectOp = exports.SelectOp || (exports.SelectOp = {}));
var Select;
(function (Select) {
    function create(ctes, body, setOps, orderBy, limit) {
        return {
            kind: 'Select',
            ctes,
            body,
            setOps,
            orderBy,
            limit,
        };
    }
    Select.create = create;
})(Select = exports.Select || (exports.Select = {}));
var Values;
(function (Values) {
    Values.defaultValues = { kind: 'DefaultValues' };
    function createExpressionValues(valuesList) {
        return { kind: 'ExpressionValues', valuesList };
    }
    Values.createExpressionValues = createExpressionValues;
    function walk(values, handlers) {
        switch (values.kind) {
            case 'DefaultValues':
                return handlers.defaultValues(values);
            case 'ExpressionValues':
                return handlers.exprValues(values);
        }
    }
    Values.walk = walk;
})(Values = exports.Values || (exports.Values = {}));
var Insert;
(function (Insert) {
    function create(ctes, table, as, columns, values, returning) {
        return {
            kind: 'Insert',
            ctes,
            table,
            as,
            columns,
            values,
            returning,
        };
    }
    Insert.create = create;
})(Insert = exports.Insert || (exports.Insert = {}));
var Update;
(function (Update) {
    function create(ctes, table, as, updates, from, where, returning) {
        return {
            kind: 'Update',
            ctes,
            table,
            as,
            updates,
            from,
            where,
            returning,
        };
    }
    Update.create = create;
})(Update = exports.Update || (exports.Update = {}));
var Delete;
(function (Delete) {
    function create(table, as, where, returning) {
        return { kind: 'Delete', table, as, where, returning };
    }
    Delete.create = create;
})(Delete = exports.Delete || (exports.Delete = {}));
var WithQuery;
(function (WithQuery) {
    function create(as, columnNames, query) {
        return { as, columnNames, query };
    }
    WithQuery.create = create;
})(WithQuery = exports.WithQuery || (exports.WithQuery = {}));
function walk(ast, handlers) {
    switch (ast.kind) {
        case 'Select':
            return handlers.select(ast);
        case 'Insert':
            return handlers.insert(ast);
        case 'Update':
            return handlers.update(ast);
        case 'Delete':
            return handlers.delete(ast);
    }
}
exports.walk = walk;
