#!/usr/bin/env node
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
// tslint:disable:no-console
const fs_1 = require("fs");
const node_watch_1 = require("node-watch");
const path = require("path");
const camelCase = require("camelcase");
const Array = require("fp-ts/lib/Array");
const Either = require("fp-ts/lib/Either");
const Option = require("fp-ts/lib/Option");
const Ord = require("fp-ts/lib/Ord");
const Ordering = require("fp-ts/lib/Ordering");
const Task = require("fp-ts/lib/Task");
const TaskEither = require("fp-ts/lib/TaskEither");
const function_1 = require("fp-ts/lib/function");
const pipeable_1 = require("fp-ts/lib/pipeable");
const yargs = require("yargs");
const clients_1 = require("./clients");
const index_1 = require("./index");
const fp_utils_1 = require("./fp-utils");
const Warn = require("./Warn");
function main() {
    return __awaiter(this, void 0, void 0, function* () {
        const args = parseArgs();
        if (!args._.length) {
            console.error('No input files. Try with `--help`.');
            return 1;
        }
        if (args.watch && args.check) {
            console.error('Cannot use --watch and --check together');
            return 1;
        }
        const options = {
            verbose: args.verbose,
            index: args.index,
            prettify: args.prettify,
            pgModule: args['pg-module'],
            terminalColumns: process.stdout.columns,
        };
        const dirPaths = [];
        for (const dirPath of args._) {
            if (!(yield fs_1.promises.stat(dirPath)).isDirectory()) {
                console.error(`Not a directory: ${dirPath}`);
                return 1;
            }
            dirPaths.push(dirPath);
        }
        const fileExtensions = extensions(args.ext);
        const clients = yield clients_1.connect(args.database);
        if (Either.isLeft(clients)) {
            console.error(clients.left);
            throw process.exit(1);
        }
        let status = 0;
        if (args.watch) {
            yield watchDirectories(clients.right, fileExtensions, dirPaths, options);
        }
        else if (args.check) {
            const result = yield checkDirectories(clients.right, fileExtensions, dirPaths, options)();
            if (!result.every(function_1.identity)) {
                console.error(`
Some files are out of date!`);
                status = 1;
            }
        }
        else {
            yield processDirectories(clients.right, fileExtensions, dirPaths, options)();
        }
        yield clients_1.disconnect(clients.right);
        return status;
    });
}
function parseArgs() {
    return yargs
        .usage('Usage: $0 [options] DIRECTORY...')
        .option('database', {
        alias: 'd',
        describe: 'Database URI to connect to, e.g. -d postgres://user:pass@localhost/mydb. ' +
            'If not given, relies node-postgres default connecting logic which uses ' +
            'environment variables',
        type: 'string',
    })
        .option('ext', {
        alias: 'e',
        describe: 'File extensions to consider, e.g. -e sql,psql',
        type: 'string',
        default: 'sql',
    })
        .option('verbose', {
        alias: 'v',
        describe: 'Give verbose output about problems with inferring statement nullability.',
        type: 'boolean',
        default: false,
    })
        .option('index', {
        describe: 'Generate an index.ts file that re-exports all generated functions',
        type: 'boolean',
        default: true,
    })
        .option('watch', {
        alias: 'w',
        description: 'Watch files and run the conversion when something changes',
        type: 'boolean',
        default: false,
    })
        .option('check', {
        alias: 'c',
        description: 'Check whether all output files are up-to-date without actually updating ' +
            'them. If they are, exit with status 0, otherwise exit with status 1. ' +
            'Useful for CI or pre-commit hooks.',
        type: 'boolean',
        default: false,
    })
        .option('prettify', {
        alias: 'p',
        description: 'Apply prettier to output TypeScript files',
        type: 'boolean',
        default: false,
    })
        .option('pg-module', {
        description: 'Where to import node-postgres from.',
        type: 'string',
        default: 'pg',
    })
        .epilogue(`\
Generate TypeScript functions for SQL statements in all files in the \
given directories. For each input file, the output file name is \
generated by removing the file extension and appending ".ts".

Each output file will export a single function whose name is a \
camelCased version of the basename of the input file.

$0 connects to the database to infer the parameter and output column \
types of each SQL statement. It does this without actually executing \
the SQL queries, so it's safe to run against any database.
`)
        .help().argv;
}
function watchDirectories(clients, fileExtensions, dirPaths, options) {
    return __awaiter(this, void 0, void 0, function* () {
        let moduleDirs = yield processDirectories(clients, fileExtensions, dirPaths, options)();
        console.log('Watching for file changes...');
        const eventBuffer = [];
        let handlingEvents = false;
        const eventHandler = makeWatchEventHandler(clients, options);
        dirPaths.forEach(dirPath => node_watch_1.default(dirPath, { filter: fileName => hasOneOfExtensions(fileExtensions, fileName) }, (event, filePath) => __awaiter(this, void 0, void 0, function* () {
            eventBuffer.push({
                type: event,
                dirPath,
                fileName: path.relative(dirPath, filePath),
            });
            if (!handlingEvents) {
                handlingEvents = true;
                moduleDirs = yield handleWatchEvents(moduleDirs, eventBuffer, eventHandler);
                handlingEvents = false;
            }
        })));
        return new Promise(() => { });
    });
}
function handleWatchEvents(moduleDirs, events, eventHandler) {
    return __awaiter(this, void 0, void 0, function* () {
        while (events.length > 0) {
            const { type, dirPath, fileName } = events.shift();
            const moduleDir = moduleDirs.find(dir => dir.dirPath === dirPath);
            if (moduleDir == null)
                return moduleDirs;
            const newModules = yield eventHandler(moduleDir.modules, type, dirPath, fileName);
            moduleDirs = pipeable_1.pipe(modifyWhere(moduleDir => moduleDir.dirPath === dirPath, moduleDir => ({ dirPath: moduleDir.dirPath, modules: newModules }), moduleDirs), Option.getOrElse(() => moduleDirs));
        }
        return moduleDirs;
    });
}
function makeWatchEventHandler(clients, options) {
    return (tsModules, type, dirPath, sqlFileName) => __awaiter(this, void 0, void 0, function* () {
        const sqlFilePath = path.join(dirPath, sqlFileName);
        let result;
        switch (type) {
            case 'update':
                result = pipeable_1.pipe(processSQLFile(clients, sqlFilePath, false, options), Task.map(tsModuleOption => pipeable_1.pipe(tsModuleOption, Option.map(tsModule => replaceOrAddTsModule(tsModule, tsModules)), Option.getOrElse(() => removeTsModule(sqlFileName, tsModules)))));
                break;
            case 'remove':
                yield removeOutputFile(sqlFilePath);
                result = pipeable_1.pipe(Task.of(removeTsModule(sqlFileName, tsModules)));
                break;
            default:
                throw new Error('never reached');
        }
        result = pipeable_1.pipe(result, Task.chain(newModules => maybeWriteIndexModule(options.index, dirPath, newModules, options.prettify)));
        return yield result();
    });
}
function replaceOrAddTsModule(tsModule, tsModules) {
    return pipeable_1.pipe(modifyWhere(mod => mod.sqlFileName === tsModule.sqlFileName, () => tsModule, tsModules), Option.getOrElse(() => Array.snoc(tsModules, tsModule)));
}
function modifyWhere(pred, replacer, where) {
    return pipeable_1.pipe(where.findIndex(pred), Option.fromNullable, Option.chain(index => pipeable_1.pipe(where, Array.modifyAt(index, () => replacer(where[index])))));
}
function removeTsModule(sqlFileName, tsModules) {
    return tsModules.filter(mod => mod.sqlFileName != sqlFileName);
}
function processDirectories(clients, fileExtensions, dirPaths, options) {
    return pipeable_1.pipe(() => __awaiter(this, void 0, void 0, function* () {
        console.log('Starting compilation...');
    }), Task.chain(() => mapDirectories(dirPaths, fileExtensions, filePath => processSQLFile(clients, filePath, false, options), (dirPath, tsModules) => processSQLDirectory(dirPath, tsModules, options))), Task.map(result => {
        console.log('done.');
        return result;
    }));
}
function checkDirectories(clients, fileExtensions, dirPaths, options) {
    return mapDirectories(dirPaths, fileExtensions, filePath => processSQLFile(clients, filePath, true, options), (_dirPath, tsModules) => checkDirectoryResult(tsModules));
}
function checkDirectoryResult(tsModules) {
    return Task.of(tsModules.every(Option.isSome));
}
function mapDirectories(dirPaths, fileExtensions, fileProcessor, dirProcessor) {
    return fp_utils_1.traverseATs(dirPaths, dirPath => mapDirectory(dirPath, fileExtensions, fileProcessor, dirProcessor));
}
function mapDirectory(dirPath, fileExtensions, fileProcessor, dirProcessor) {
    return pipeable_1.pipe(findSQLFilePaths(dirPath, fileExtensions), Task.chain(filePaths => fp_utils_1.traverseATs(filePaths, fileProcessor)), Task.chain(results => dirProcessor(dirPath, results)));
}
function processSQLFile(clients, filePath, checkOnly, options) {
    const tsPath = getOutputPath(filePath);
    const fnName = funcName(filePath);
    return pipeable_1.pipe(() => fs_1.promises.readFile(filePath, 'utf-8'), Task.chain(sql => index_1.sqlToStatementDescription(clients, sql)), TaskEither.map(stmtWithWarnings => {
        const [stmt, warnings] = Warn.split(stmtWithWarnings);
        if (warnings.length > 0) {
            console.warn(Warn.format(warnings, options.verbose, options.terminalColumns || 78));
        }
        return stmt;
    }), TaskEither.chain(source => index_1.generateTSCode(clients, path.basename(filePath), source, fnName, {
        prettierFileName: options.prettify ? tsPath : undefined,
        pgModule: options.pgModule,
    })), TaskEither.chain(tsCode => () => __awaiter(this, void 0, void 0, function* () {
        if (yield isFileOutOfDate(tsPath, tsCode)) {
            if (checkOnly) {
                return Either.left('out of date');
            }
            console.log(`Writing ${tsPath}`);
            yield fs_1.promises.writeFile(tsPath, tsCode).then(Either.right);
        }
        return Either.right(undefined);
    })), TaskEither.orElse(errorMessage => () => __awaiter(this, void 0, void 0, function* () {
        console.error(`${filePath}: ${errorMessage}`);
        if (!checkOnly) {
            console.log(`Removing ${tsPath}`);
            yield tryCatch(() => fs_1.promises.unlink(tsPath));
        }
        return Either.left(undefined);
    })), TaskEither.map(() => ({
        sqlFileName: path.basename(filePath),
        tsFileName: path.basename(tsPath),
        funcName: fnName,
    })), Task.map(Option.fromEither));
}
function processSQLDirectory(dirPath, modules, options) {
    const successfulModules = pipeable_1.pipe(modules, Array.filterMap(function_1.identity));
    return pipeable_1.pipe(maybeWriteIndexModule(options.index, dirPath, successfulModules, options.prettify), Task.map(modules => ({ dirPath, modules })));
}
function maybeWriteIndexModule(write, dirPath, tsModules, prettify) {
    const tsPath = path.join(dirPath, 'index.ts');
    if (write) {
        return pipeable_1.pipe(Task.of(tsModules), Task.map(modules => pipeable_1.pipe(modules, Array.sort(Ord.fromCompare((a, b) => Ordering.sign(a.tsFileName.localeCompare(b.tsFileName)))))), Task.chain(sortedModules => index_1.indexModuleTS(sortedModules, {
            prettierFileName: prettify ? tsPath : null,
        })), Task.chain(tsCode => () => __awaiter(this, void 0, void 0, function* () {
            if (yield isFileOutOfDate(tsPath, tsCode)) {
                console.log(`Writing ${tsPath}`);
                yield fs_1.promises.writeFile(tsPath, tsCode);
            }
        })), Task.map(() => tsModules));
    }
    return Task.of(tsModules);
}
function isFileOutOfDate(filePath, newContents) {
    return __awaiter(this, void 0, void 0, function* () {
        let oldContents;
        try {
            oldContents = yield fs_1.promises.readFile(filePath, 'utf-8');
        }
        catch (_err) {
            oldContents = null;
        }
        return oldContents !== newContents;
    });
}
function tryCatch(f) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            return yield f();
        }
        catch (_err) {
            return null;
        }
    });
}
function funcName(filePath) {
    const parsed = path.parse(filePath);
    return camelCase(parsed.name);
}
function removeOutputFile(filePath) {
    return __awaiter(this, void 0, void 0, function* () {
        const tsPath = getOutputPath(filePath);
        try {
            yield fs_1.promises.unlink(tsPath);
        }
        catch (_err) {
            return;
        }
        console.log(`Removing ${tsPath}`);
    });
}
function findSQLFilePaths(dirPath, fileExtensions) {
    return pipeable_1.pipe(() => fs_1.promises.readdir(dirPath, {
        encoding: 'utf-8',
        withFileTypes: true,
    }), Task.chain(dirents => pipeable_1.pipe(fp_utils_1.traverseATs(dirents, dirent => pipeable_1.pipe(isSQLFile(fileExtensions, dirPath, dirent.name), Task.map(is => (is ? Option.some(dirent) : Option.none)))), Task.map(Array.filterMap(function_1.identity)), Task.map(dirents => dirents.map(dirent => path.join(dirPath, dirent.name))))));
}
function getOutputPath(filePath) {
    return path.format(Object.assign(Object.assign({}, path.parse(filePath)), { base: undefined, ext: '.ts' }));
}
function extensions(e) {
    return e.split(',').map(ext => `.${ext}`);
}
function isSQLFile(extensions, dirPath, fileName) {
    return () => __awaiter(this, void 0, void 0, function* () {
        let stats;
        try {
            stats = yield fs_1.promises.stat(path.join(dirPath, fileName));
        }
        catch (_err) {
            return false;
        }
        return stats.isFile() && hasOneOfExtensions(extensions, fileName);
    });
}
function hasOneOfExtensions(exts, fileName) {
    return exts.includes(path.parse(fileName).ext);
}
main()
    .then(status => process.exit(status))
    .catch(err => {
    console.error(err);
    process.exit(99);
});
