"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const Either = require("fp-ts/lib/Either");
const Task = require("fp-ts/lib/Task");
const pipeable_1 = require("fp-ts/lib/pipeable");
const fp_utils_1 = require("./fp-utils");
const pkginfo = require("./pkginfo");
////////////////////////////////////////////////////////////////////////
function validateStatement(stmt) {
    const columnNames = new Set();
    const conflicts = new Set();
    stmt.columns.forEach(({ name }) => {
        if (columnNames.has(name)) {
            conflicts.add(name);
        }
        else {
            columnNames.add(name);
        }
    });
    if (conflicts.size) {
        const dup = [...conflicts.values()].sort().join(', ');
        return Either.left(`Duplicate output columns: ${dup}`);
    }
    return Either.right(stmt);
}
exports.validateStatement = validateStatement;
////////////////////////////////////////////////////////////////////////
function generateTypeScript(types, sourceFileName, pgModule, funcName, stmt) {
    const positionalOnly = hasOnlyPositionalParams(stmt);
    return pipeable_1.pipe(Task.of(typeScriptString(sourceFileName, pgModule, funcName, stmt.sql)), Task.ap(funcParams(types, stmt, positionalOnly)), Task.ap(funcReturnType(types, stmt)), Task.ap(Task.of(queryValues(stmt, positionalOnly))), Task.ap(Task.of(outputValue(stmt))));
}
exports.generateTypeScript = generateTypeScript;
const typeScriptString = (sourceFileName, pgModule, funcName, sql) => (params) => (returnType) => (queryValues) => (outputValue) => `\
// Generated by ${pkginfo.name} ${pkginfo.version} from ${sourceFileName}.
// Do not edit directly. Instead, edit ${sourceFileName} and re-run ${pkginfo.name}.

import { ClientBase } from '${pgModule}'

export async function ${funcName}(
  client: ClientBase${params}
): Promise<${returnType}> {
    const result = await client.query(\`\\
${sql}\`${queryValues})
    return ${outputValue}
}
`;
function hasOnlyPositionalParams(stmt) {
    return stmt.params.every(param => !!param.name.match(/\$\d+/));
}
function funcReturnType(types, stmt) {
    return pipeable_1.pipe(fp_utils_1.traverseATs(stmt.columns, columnType(types)), Task.map(columnTypes => {
        const rowType = '{ ' + columnTypes.join('; ') + ' }';
        switch (stmt.rowCount) {
            case 'zero':
                return 'number'; // return the affected row count
            case 'one':
                return rowType;
            case 'zeroOrOne':
                return `${rowType} | null`;
            case 'many':
                return `Array<${rowType}>`;
        }
    }));
}
const columnType = (types) => (column) => {
    return pipeable_1.pipe(types.columnType(column), Task.map(({ name, type }) => `${stringLiteral(name)}: ${type}`));
};
function outputValue(stmt) {
    switch (stmt.rowCount) {
        case 'zero':
            return 'result.rowCount'; // return the affected row count
        case 'one':
            return 'result.rows[0]';
        case 'zeroOrOne':
            return 'result.rows.length ? result.rows[0] : null';
        case 'many':
            return 'result.rows';
    }
}
function stringLiteral(str) {
    return "'" + str.replace('\\', '\\\\').replace("'", "\\'") + "'";
}
function funcParams(types, stmt, positionalOnly) {
    if (!stmt.params.length) {
        return Task.of('');
    }
    return pipeable_1.pipe(positionalOnly
        ? positionalFuncParams(types, stmt)
        : namedFuncParams(types, stmt), Task.map(params => `, ${params}`));
}
function positionalFuncParams(types, stmt) {
    return pipeable_1.pipe(fp_utils_1.traverseATs(stmt.params, param => pipeable_1.pipe(types.tsType(param.type, param.nullable), Task.map(tsType => `${param.name}: ${tsType}`))), Task.map(params => params.join(', ')));
}
function namedFuncParams(types, stmt) {
    return pipeable_1.pipe(positionalFuncParams(types, stmt), Task.map(params => `params: { ${params} }`));
}
function queryValues(stmt, positionalOnly) {
    if (!stmt.params.length) {
        return '';
    }
    const prefix = positionalOnly ? '' : 'params.';
    return (', [ ' + stmt.params.map(param => prefix + param.name).join(', ') + ' ]');
}
function generateIndexModule(modules) {
    return modules
        .map(({ tsFileName, funcName }) => `export { ${funcName} } from './${baseNameWithoutExt(tsFileName)}';`)
        .join('\n');
}
exports.generateIndexModule = generateIndexModule;
function baseNameWithoutExt(filePath) {
    return path.parse(filePath).name;
}
