"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Either_1 = require("fp-ts/lib/Either");
const typed_parser_1 = require("typed-parser");
const ast_1 = require("../ast");
const utils_1 = require("./utils");
const token_1 = require("./token");
const special_functions_1 = require("./special-functions");
// ( ... )
function parenthesized(parser) {
    return typed_parser_1.seq(typed_parser_1.$2, token_1.symbol('('), parser, token_1.symbol(')'));
}
exports.parenthesized = parenthesized;
// [ AS ] identifier
const as = typed_parser_1.seq((_as, id) => id, utils_1.optional(token_1.reservedWord('AS')), token_1.identifier);
// AS identifier
const reqAs = typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('AS'), token_1.identifier);
// [ schema . ] table
const tableRef = typed_parser_1.seq((id1, id2) => (id2 ? ast_1.TableRef.create(id1, id2) : ast_1.TableRef.create(null, id1)), token_1.identifier, utils_1.optional(typed_parser_1.seq(typed_parser_1.$2, token_1.symbol('.'), token_1.identifier)));
// Expressions
const arraySubQueryExpr = typed_parser_1.seq((_arr, subquery) => ast_1.Expression.createArraySubQuery(subquery), token_1.reservedWord('ARRAY'), parenthesized(typed_parser_1.lazy(() => select)));
const caseBranch = typed_parser_1.seq((_when, condition, _then, result) => ({ condition, result }), token_1.reservedWord('WHEN'), typed_parser_1.lazy(() => expression), token_1.reservedWord('THEN'), typed_parser_1.lazy(() => expression));
const caseElse = typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('ELSE'), typed_parser_1.lazy(() => expression));
const caseExpr = typed_parser_1.seq((_case, branch1, branches, else_, _end) => ast_1.Expression.createCase([branch1, ...branches], else_), token_1.reservedWord('CASE'), caseBranch, typed_parser_1.many(caseBranch), utils_1.optional(caseElse), token_1.reservedWord('END'));
const functionArguments = parenthesized(typed_parser_1.oneOf(
// func(*} means no arguments
typed_parser_1.seq(_ => [], token_1.symbol('*')), typed_parser_1.sepBy(token_1.symbol(','), typed_parser_1.lazy(() => expression))));
const columnRefOrFunctionCallExpr = typed_parser_1.seq((ident, rest) => rest == null
    ? ast_1.Expression.createColumnRef(ident)
    : typeof rest === 'string'
        ? ast_1.Expression.createTableColumnRef(ident, rest)
        : ast_1.Expression.createFunctionCall(ident, rest), token_1.identifier, typed_parser_1.oneOf(typed_parser_1.seq(typed_parser_1.$2, token_1.symbol('.'), token_1.identifier), functionArguments, token_1._));
const strEscape = typed_parser_1.seq(typed_parser_1.$2, token_1.symbolKeepWS('\\'), typed_parser_1.oneOf(typed_parser_1.keyword("'", "'"), typed_parser_1.keyword('\\', '\\'), typed_parser_1.keyword('/', '/'), typed_parser_1.keyword('b', '\b'), typed_parser_1.keyword('f', '\f'), typed_parser_1.keyword('n', '\n'), typed_parser_1.keyword('r', '\r'), typed_parser_1.keyword('t', '\t')));
const strInner = typed_parser_1.seq((s, tail) => s + tail, typed_parser_1.stringBefore("[\\']"), typed_parser_1.oneOf(typed_parser_1.seq((e, t) => e + t, strEscape, typed_parser_1.lazy(() => strInner)), typed_parser_1.constant('')));
const stringConstant = typed_parser_1.seq(typed_parser_1.$2, token_1.symbolKeepWS("'"), strInner, token_1.symbol("'"));
const constantExpr = typed_parser_1.seq((val, _ws) => ast_1.Expression.createConstant(val), typed_parser_1.oneOf(token_1.expectIdentifier('TRUE'), token_1.expectIdentifier('FALSE'), token_1.expectIdentifier('NULL'), typed_parser_1.match('[0-9]+(\\.[0-9]+)?([eE]-?[0-9]+)?'), typed_parser_1.match('\\.[0-9]+'), stringConstant), token_1._);
const parameterExpr = typed_parser_1.seq((_$, index, _ws) => ast_1.Expression.createParameter(index), token_1.symbolKeepWS('$'), typed_parser_1.int('[0-9]+'), token_1._);
const parenthesizedExpr = parenthesized(typed_parser_1.lazy(() => expression));
const typeName = typed_parser_1.seq((id, arraySuffix) => id + arraySuffix, token_1.identifier, typed_parser_1.oneOf(typed_parser_1.seq(_ => '[]', token_1.symbol('['), token_1.symbol(']')), typed_parser_1.seq(_ => '', token_1._)));
const primaryExpr = typed_parser_1.seq((expr, typeCast) => typeCast != null ? ast_1.Expression.createTypeCast(expr, typeCast) : expr, typed_parser_1.oneOf(arraySubQueryExpr, caseExpr, typed_parser_1.attempt(special_functions_1.specialFunctionCall(typed_parser_1.lazy(() => primaryExpr))), columnRefOrFunctionCallExpr, constantExpr, parameterExpr, parenthesizedExpr), utils_1.optional(typed_parser_1.seq(typed_parser_1.$2, token_1.symbol('::'), typeName)));
function makeUnaryOp(oper, nextExpr) {
    return typed_parser_1.seq((ops, next) => ops.length > 0
        ? ops.reduceRight((acc, op) => ast_1.Expression.createUnaryOp(op, acc), next)
        : next, typed_parser_1.many(oper), nextExpr);
}
function makeBinaryOp(oper, nextExpr) {
    return typed_parser_1.seq((first, rest) => rest.reduce((acc, val) => ast_1.Expression.createBinaryOp(acc, val.op, val.next), first), nextExpr, typed_parser_1.many(typed_parser_1.seq((op, next) => ({ op, next }), oper, nextExpr)));
}
const oneOfOperators = (...ops) => typed_parser_1.oneOf(...ops.map(token_1.operator));
const subscriptExpr = typed_parser_1.seq((next, subs) => subs.reduce((acc, val) => ast_1.Expression.createBinaryOp(acc, '[]', val), next), primaryExpr, typed_parser_1.many(typed_parser_1.seq(typed_parser_1.$2, token_1.symbol('['), typed_parser_1.lazy(() => expression), token_1.symbol(']'))));
const unaryPlusMinus = makeUnaryOp(oneOfOperators('+', '-'), subscriptExpr);
const expExpr = makeBinaryOp(token_1.operator('^'), unaryPlusMinus);
const mulDivModExpr = makeBinaryOp(oneOfOperators('*', '/', '%'), expExpr);
const addSubExpr = makeBinaryOp(oneOfOperators('+', '-'), mulDivModExpr);
const existsExpr = typed_parser_1.seq((_exists, subquery) => ast_1.Expression.createExistsOp(subquery), token_1.reservedWord('EXISTS'), parenthesized(typed_parser_1.lazy(() => select)));
var OtherExprRhs;
(function (OtherExprRhs) {
    const in_ = typed_parser_1.seq((op, rhs) => ({ kind: 'InExprRhs', op, rhs }), typed_parser_1.attempt(typed_parser_1.oneOf(token_1.reservedWord('IN'), typed_parser_1.seq(_ => 'NOT IN', token_1.reservedWord('NOT'), token_1.reservedWord('IN')))), parenthesized(typed_parser_1.lazy(() => select)));
    const ternary = typed_parser_1.seq((op, rhs1, _and, rhs2) => ({ kind: 'TernaryExprRhs', op, rhs1, rhs2 }), typed_parser_1.oneOf(typed_parser_1.attempt(token_1.sepReserveds('NOT BETWEEN SYMMETRIC')), typed_parser_1.attempt(token_1.sepReserveds('BETWEEN SYMMETRIC')), typed_parser_1.attempt(token_1.sepReserveds('NOT BETWEEN')), typed_parser_1.attempt(token_1.reservedWord('BETWEEN'))), addSubExpr, token_1.reservedWord('AND'), addSubExpr);
    const unarySuffix = typed_parser_1.seq(_ => ({ kind: 'UnarySuffix', op: '!' }), token_1.operator('!'));
    const otherOp = typed_parser_1.seq((op, rhs) => ({ kind: 'OtherOpExprRhs', op, rhs }), typed_parser_1.oneOf(token_1.anyOperatorExcept([
        '<',
        '>',
        '=',
        '<=',
        '>=',
        '<>',
        '+',
        '-',
        '*',
        '/',
        '%',
        '^',
    ]), token_1.sepReserveds('IS DISTINCT FROM'), token_1.sepReserveds('IS NOT DISTINCT FROM'), token_1.reservedWord('LIKE')), addSubExpr);
    OtherExprRhs.parser = typed_parser_1.oneOf(in_, ternary, unarySuffix, otherOp);
    function createExpression(lhs, rhs) {
        switch (rhs.kind) {
            case 'InExprRhs':
                return ast_1.Expression.createInOp(lhs, rhs.op, rhs.rhs);
            case 'TernaryExprRhs':
                return ast_1.Expression.createTernaryOp(lhs, rhs.op, rhs.rhs1, rhs.rhs2);
            case 'UnarySuffix':
                return ast_1.Expression.createUnaryOp(rhs.op, lhs);
            case 'OtherOpExprRhs':
                return ast_1.Expression.createBinaryOp(lhs, rhs.op, rhs.rhs);
        }
    }
    OtherExprRhs.createExpression = createExpression;
})(OtherExprRhs || (OtherExprRhs = {}));
const otherExpr = typed_parser_1.seq((first, rest) => rest.reduce((acc, val) => OtherExprRhs.createExpression(acc, val), first), addSubExpr, typed_parser_1.many(OtherExprRhs.parser));
const existsOrOtherExpr = typed_parser_1.oneOf(existsExpr, otherExpr);
const comparisonExpr = makeBinaryOp(oneOfOperators('<', '>', '=', '<=', '>=', '<>'), existsOrOtherExpr);
const isExpr = typed_parser_1.seq((next, op) => (op ? ast_1.Expression.createUnaryOp(op, next) : next), comparisonExpr, utils_1.optional(typed_parser_1.oneOf(token_1.sepReserveds('IS NULL'), token_1.sepReserveds('IS NOT NULL'), token_1.reservedWord('ISNULL'), token_1.reservedWord('NOTNULL'), token_1.sepReserveds('IS TRUE'), token_1.sepReserveds('IS NOT TRUE'), token_1.sepReserveds('IS FALSE'), token_1.sepReserveds('IS NOT FALSE'), token_1.sepReserveds('IS UNKNOWN'), token_1.sepReserveds('IS NOT UNKNOWN'))));
const notExpr = makeUnaryOp(token_1.reservedWord('NOT'), isExpr);
const andExpr = makeBinaryOp(token_1.reservedWord('AND'), notExpr);
const orExpr = makeBinaryOp(token_1.reservedWord('OR'), andExpr);
const expression = orExpr;
// (name1, name2, ...)
const identifierList = parenthesized(typed_parser_1.sepBy1(token_1.symbol(','), token_1.identifier));
// WITH
const withQueries = typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('WITH'), typed_parser_1.sepBy1(token_1.symbol(','), typed_parser_1.seq((as, columns, _as, stmt) => ast_1.WithQuery.create(as, columns, stmt), token_1.identifier, utils_1.optional(identifierList), token_1.reservedWord('AS'), parenthesized(typed_parser_1.lazy(() => statementParser)))));
const crossJoin = typed_parser_1.seq((_cj, tableExpr) => ({ join: { kind: 'Cross' }, tableExpr }), token_1.sepReserveds('CROSS JOIN'), typed_parser_1.lazy(() => tableExpression));
const qualifiedJoinType = typed_parser_1.seq((joinType, _join) => joinType || 'INNER', utils_1.optional(typed_parser_1.oneOf(token_1.reservedWord('INNER'), typed_parser_1.seq(typed_parser_1.$1, typed_parser_1.oneOf(token_1.reservedWord('LEFT'), token_1.reservedWord('RIGHT'), token_1.reservedWord('FULL')), utils_1.optional(token_1.reservedWord('OUTER'))))), token_1.reservedWord('JOIN'));
const qualifiedJoin = typed_parser_1.seq((type, tableExpr, _on, condition) => ({
    join: { kind: 'Qualified', type, condition },
    tableExpr,
}), qualifiedJoinType, typed_parser_1.lazy(() => tableExpression), token_1.reservedWord('ON'), expression);
const naturalJoinType = typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('NATURAL'), qualifiedJoinType);
const naturalJoin = typed_parser_1.seq((type, tableExpr) => ({ join: { kind: 'Natural', type }, tableExpr }), naturalJoinType, typed_parser_1.lazy(() => tableExpression));
const table = typed_parser_1.seq(ast_1.TableExpression.createTable, tableRef, utils_1.optional(as));
function tableExprReducer(acc, joinSpec) {
    switch (joinSpec.join.kind) {
        case 'Cross':
            return ast_1.TableExpression.createCrossJoin(acc, joinSpec.tableExpr);
        case 'Qualified':
            return ast_1.TableExpression.createQualifiedJoin(acc, joinSpec.join.type, joinSpec.tableExpr, joinSpec.join.condition);
        case 'Natural':
            return ast_1.TableExpression.createQualifiedJoin(acc, joinSpec.join.type, joinSpec.tableExpr, null // null conditiong means NATURAL JOIN
            );
    }
}
const tableExpression = typed_parser_1.seq((lhs, rest) => (rest.length === 0 ? lhs : rest.reduce(tableExprReducer, lhs)), typed_parser_1.oneOf(typed_parser_1.attempt(typed_parser_1.seq(typed_parser_1.$2, token_1.symbol('('), typed_parser_1.lazy(() => tableExpression), token_1.symbol(')'))), typed_parser_1.attempt(typed_parser_1.seq((stmt, as) => ast_1.TableExpression.createSubQuery(stmt, as), parenthesized(typed_parser_1.lazy(() => statementParser)), as)), table), typed_parser_1.many(typed_parser_1.oneOf(crossJoin, qualifiedJoin, naturalJoin)));
const from = typed_parser_1.seq((_from, tableExpr, rest) => rest.length === 0
    ? tableExpr
    : // Implicit join equals to CROSS JOIN
        rest.reduce((acc, next) => ast_1.TableExpression.createCrossJoin(acc, next), tableExpr), token_1.reservedWord('FROM'), tableExpression, typed_parser_1.many(typed_parser_1.seq(typed_parser_1.$2, token_1.symbol(','), tableExpression)));
// WHERE
const where = typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('WHERE'), expression);
// GROUP BY & HAVING
const groupBy = typed_parser_1.seq(typed_parser_1.$3, token_1.reservedWord('GROUP'), token_1.reservedWord('BY'), typed_parser_1.sepBy1(token_1.symbol(','), expression));
const having = typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('HAVING'), expression);
// ORDER BY
const orderByOrder = typed_parser_1.oneOf(token_1.reservedWord('ASC'), token_1.reservedWord('DESC'), typed_parser_1.seq((_using, op) => ['USING', op], token_1.reservedWord('USING'), token_1.anyOperator));
const orderByNulls = typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('NULLS'), typed_parser_1.oneOf(token_1.reservedWord('FIRST'), token_1.reservedWord('LAST')));
const orderByItem = typed_parser_1.seq((expr, order, nulls) => ast_1.OrderBy.create(expr, order, nulls), expression, utils_1.optional(orderByOrder), utils_1.optional(orderByNulls));
const orderBy = typed_parser_1.seq((_order, _by, list) => list, token_1.reservedWord('ORDER'), token_1.reservedWord('BY'), typed_parser_1.sepBy1(token_1.symbol(','), orderByItem));
// LIMIT
const limit = typed_parser_1.seq((_l, count, offset) => ast_1.Limit.create(count, offset), token_1.reservedWord('LIMIT'), typed_parser_1.oneOf(typed_parser_1.seq(typed_parser_1.$null, token_1.reservedWord('ALL')), expression), utils_1.optional(typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('OFFSET'), expression)));
// SELECT
const allFields = typed_parser_1.seq(_ => ast_1.SelectListItem.createAllFields(), token_1.symbol('*'));
const allTableFields = typed_parser_1.seq((table, _p, _a) => ast_1.SelectListItem.createAllTableFields(table), token_1.identifier, token_1.symbol('.'), token_1.symbol('*'));
const selectListExpression = typed_parser_1.seq((expr, as) => ast_1.SelectListItem.createSelectListExpression(expr, as), expression, utils_1.optional(as));
const selectListItem = typed_parser_1.oneOf(allFields, typed_parser_1.attempt(allTableFields), selectListExpression);
const selectList = typed_parser_1.sepBy1(token_1.symbol(','), selectListItem);
const selectBody = typed_parser_1.seq((_sel, list, from, where, groupBy, having) => ast_1.SelectBody.create(list, from, where, groupBy || [], having), token_1.reservedWord('SELECT'), selectList, utils_1.optional(from), utils_1.optional(where), utils_1.optional(groupBy), utils_1.optional(having));
const selectSetOps = typed_parser_1.many(typed_parser_1.seq((op, duplicates, body) => ast_1.SelectOp.create(op, duplicates || 'DISTINCT', body), typed_parser_1.oneOf(token_1.reservedWord('UNION'), token_1.reservedWord('INTERSECT'), token_1.reservedWord('EXCEPT')), utils_1.optional(typed_parser_1.oneOf(token_1.reservedWord('ALL'), token_1.reservedWord('DISTINCT'))), selectBody));
const select = typed_parser_1.seq((withQueries, body, setOps, orderBy, limit) => ast_1.Select.create(withQueries || [], body, setOps, orderBy || [], limit), utils_1.optional(withQueries), selectBody, selectSetOps, utils_1.optional(orderBy), utils_1.optional(limit));
// INSERT
const defaultValues = typed_parser_1.seq((_def, _val) => ast_1.Values.defaultValues, token_1.reservedWord('DEFAULT'), token_1.reservedWord('VALUES'));
const expressionValues = typed_parser_1.seq((_val, values) => ast_1.Values.createExpressionValues(values), token_1.reservedWord('VALUES'), typed_parser_1.sepBy(token_1.symbol(','), parenthesized(typed_parser_1.sepBy1(token_1.symbol(','), typed_parser_1.oneOf(typed_parser_1.seq(typed_parser_1.$null, token_1.reservedWord('DEFAULT')), expression)))));
const values = typed_parser_1.oneOf(defaultValues, expressionValues);
const returning = typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('RETURNING'), selectList);
const insertInto = typed_parser_1.seq(typed_parser_1.$3, token_1.reservedWord('INSERT'), token_1.reservedWord('INTO'), tableRef);
const insert = typed_parser_1.seq((withQueries, table, as, columns, values, returning) => ast_1.Insert.create(withQueries || [], table, as, columns || [], values, returning || []), utils_1.optional(withQueries), insertInto, utils_1.optional(reqAs), utils_1.optional(identifierList), values, utils_1.optional(returning));
// UPDATE
const updateAssignments = typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('SET'), typed_parser_1.sepBy1(token_1.symbol(','), typed_parser_1.seq((columnName, _eq, value) => ({ columnName, value }), token_1.identifier, token_1.symbol('='), expression)));
const updateTable = typed_parser_1.seq(typed_parser_1.$2, token_1.reservedWord('UPDATE'), tableRef);
const update = typed_parser_1.seq((withQueries, table, as, updates, from, where, returning) => ast_1.Update.create(withQueries || [], table, as, updates, from, where, returning || []), utils_1.optional(withQueries), updateTable, utils_1.optional(reqAs), updateAssignments, utils_1.optional(from), utils_1.optional(where), utils_1.optional(returning));
// DELETE
const deleteFrom = typed_parser_1.seq(typed_parser_1.$3, token_1.reservedWord('DELETE'), token_1.reservedWord('FROM'), tableRef);
const delete_ = typed_parser_1.seq((table, as, where, returning) => ast_1.Delete.create(table, as, where, returning || []), deleteFrom, utils_1.optional(reqAs), utils_1.optional(where), utils_1.optional(returning));
// parse
const statementParser = typed_parser_1.oneOf(select, insert, update, delete_);
const topLevelParser = typed_parser_1.seq(typed_parser_1.$2, token_1._, statementParser, typed_parser_1.end);
function parse(source) {
    return Either_1.tryCatch(() => typed_parser_1.run(topLevelParser, source), e => e.explain());
}
exports.parse = parse;
