"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const Option = require("fp-ts/lib/Option");
const Task = require("fp-ts/lib/Task");
const pipeable_1 = require("fp-ts/lib/pipeable");
const types_1 = require("./types");
const schema_1 = require("./schema");
function typeClient(pgClient) {
    return __awaiter(this, void 0, void 0, function* () {
        let arrayTypes = null;
        let enums = null;
        function valueTsType(oid) {
            return exports.nodePgBuiltinTypes.get(oid) || enums.get(oid) || exports.defaultType;
        }
        function arrayTsType(oid, elemNullable) {
            if (!exports.nodePgBuiltinArrayTypes.get(oid)) {
                // node-postgres won't convert this value to an array
                return Option.none;
            }
            const elemOid = arrayTypes.get(oid);
            if (!elemOid) {
                // This type is not an array according to
                // makeArrayTypesMap. Should not happen.
                return Option.none;
            }
            const elemType = valueTsType(elemOid) + (elemNullable ? ' | null' : '');
            return Option.some(`Array<${elemType}>`);
        }
        function tsType(valueType, nullable) {
            return () => __awaiter(this, void 0, void 0, function* () {
                if (arrayTypes == null)
                    arrayTypes = yield makeArrayTypesMap(pgClient);
                if (enums == null)
                    enums = yield makeEnumMap(pgClient);
                const result = types_1.ValueType.walk(valueType, {
                    array: ({ oid, elemNullable }) => pipeable_1.pipe(arrayTsType(oid, elemNullable), Option.getOrElse(() => exports.defaultType)),
                    any: ({ oid }) => pipeable_1.pipe(
                    // If it's an array type its items can be NULL, unless we
                    // know how the array was constructed (see the array handler
                    // above).
                    arrayTsType(oid, true), Option.getOrElse(() => valueTsType(oid))),
                });
                return nullable ? `${result} | null` : result;
            });
        }
        function columnType(column) {
            return pipeable_1.pipe(tsType(column.type, column.nullable), Task.map(type => ({
                name: column.name,
                type,
            })));
        }
        return {
            tsType,
            columnType,
        };
    });
}
exports.typeClient = typeClient;
// Equals to default type conversion of node-postgres:
// https://github.com/brianc/node-pg-types/blob/master/lib/textParsers.js
//
// Note that array types have been removed, since we handle them
// specially.
//
exports.nodePgBuiltinTypes = new Map([
    [20, 'string'],
    [21, 'number'],
    [23, 'number'],
    [26, 'number'],
    [700, 'number'],
    [701, 'number'],
    [16, 'boolean'],
    [1082, 'Date'],
    [1114, 'Date'],
    [1184, 'Date'],
    [600, '{ x: number; y: number }'],
    [718, '{ x: number; y: number; radius: number }'],
    [1186, '{ hours: number; minutes: number, seconds: number }'],
    [17, 'Buffer'],
    // TODO: JSON could be typed more accurately because it only has
    // string, number, boolean, null, and objects and arrays of them
    [114, 'any'],
    [3802, 'any'],
]);
exports.nodePgBuiltinArrayTypes = new Map([
    [651, 'string[]'],
    [1000, 'boolean[]'],
    [1001, 'Buffer[]'],
    [1005, 'number[]'],
    [1007, 'number[]'],
    [1028, 'number[]'],
    [1016, 'string[]'],
    [1017, '{ x: number; y: number }[]'],
    [1021, 'number[]'],
    [1022, 'number[]'],
    [1231, 'string[]'],
    [1014, 'string[]'],
    [1015, 'string[]'],
    [1008, 'string[]'],
    [1009, 'string[]'],
    [1040, 'string[]'],
    [1041, 'string[]'],
    [1115, 'Date[]'],
    [1182, 'Date[]'],
    [1185, 'Date[]'],
    [1187, '{ hours: number; minutes: number, seconds: number }[]'],
    [199, 'any[]'],
    [3807, 'any[]'],
    [3907, 'string[]'],
    [2951, 'string[]'],
    [791, 'string[]'],
    [1183, 'string[]'],
    [1270, 'string[]'],
]);
exports.defaultType = 'string';
function makeArrayTypesMap(pgClient) {
    return __awaiter(this, void 0, void 0, function* () {
        const arrayTypes = yield schema_1.schemaClient(pgClient).getArrayTypes();
        return new Map(arrayTypes.map(({ oid, elemType }) => [oid, elemType]));
    });
}
function makeEnumMap(pgClient) {
    return __awaiter(this, void 0, void 0, function* () {
        const enums = yield schema_1.schemaClient(pgClient).getEnums();
        return new Map(enums.map(enumType => [
            enumType.oid,
            tsUnion(enumType.labels.map(tsStringLiteral)),
        ]));
    });
}
function tsUnion(arr) {
    return arr.join(' | ');
}
function tsStringLiteral(str) {
    return ts
        .createPrinter()
        .printNode(ts.EmitHint.Expression, ts.createStringLiteral(str), ts.createSourceFile('', '', ts.ScriptTarget.Latest));
}
